import json
import logging
from typing import Any

from beet import ListOption
from pydantic import Field, root_validator, validator
from weld.merging.parser import get

from .base import BaseModel
from .priority import Priority
from .rules import AdditiveRule, Rule
from .sources import ReferenceSource, ValueSource

logger = logging.getLogger(__name__)


def deserialize(model: BaseModel, defaults: bool = True):
    return json.loads(model.json(by_alias=True, exclude_defaults=not defaults))


class SmithedModel(BaseModel, extra="forbid"):
    id: str = ""
    version: int = 1
    override: bool = False  # only should be set by bundle packs
    rules: list[Rule] = []
    priority: Priority | None = None

    @validator("priority", always=True)
    def push_down_priorities(cls, value: Priority, values: dict[str, list[Rule]]):
        """Push down top-level priority to every rule.

        If a rule has a priority defined, it will not be overwritten.
        """

        if value is None:
            return Priority()

        for rule in values["rules"]:
            if rule.priority is None:
                rule.priority = value


class SmithedJsonFile(BaseModel, extra="allow"):
    """Accepts any standard JSON file from in-game, only needs as `__smithed__` field
    for custom merging logic"""

    smithed: ListOption[SmithedModel] = Field(
        default_factory=ListOption, alias="__smithed__"
    )

    @root_validator
    def convert_type(cls, values: dict[str, ListOption[SmithedModel]]):
        if smithed := values.get("smithed"):
            for model in smithed.entries():
                model.rules = list(cls.convert_rules(model.rules, values))

        return values

    @classmethod
    def convert_rules(cls, rules: list["Rule"], values: dict[str, Any]):
        """Converts the source field of additive rules to a value source if it is
        a reference source, essentially "baking" it in for ease of use later.
        """

        for rule in rules:
            match isinstance(rule, AdditiveRule) and rule.source:
                case ReferenceSource(path=path):
                    try:
                        rule.source = ValueSource(value=get(values, path))
                    except ValueError:
                        logger.warn(
                            f"Source Reference Path: {path} was not found, deleting."
                        )
                        continue
            yield rule



from . import merging
from .toolchain import cli, inject_pack_id_into_smithed, print_pack_name, run_weld

__version__ = "0.13.1"

__all__ = [
    "print_pack_name",
    "inject_pack_id_into_smithed",
    "cli",
    "run_weld",
    "merging",
]



import logging
import tempfile
import time
from importlib import resources
from io import StringIO
from pathlib import Path
from zipfile import ZipFile

import streamlit as st
import weld
import yaml
from streamlit.delta_generator import DeltaGenerator
from streamlit_extras.stateful_button import button as toggle_button

from .model import WebApp

icon = "https://github.com/Smithed-MC/smithed-python/blob/feat/weld-but-good/smithed/weld/resources/icon.png?raw=true"

logging.basicConfig(format="%(levelname)-8s %(message)s", level=logging.INFO)

webapp = WebApp.parse_obj(
    yaml.safe_load((resources.files("weld") / "resources" / "webapp.yaml").read_text())
)
temp_dir = tempfile.TemporaryDirectory()
temp = Path(temp_dir.name)

console = logging.StreamHandler(stream := StringIO())
logging.getLogger("weld").addHandler(console)


# def validate_zips(packs: list[str]):
#     for pack in packs:
#         try:
#             with pack.open("pack.mcmeta") as meta:
#                 print(meta.read(), "\n")
#         except FileNotFoundError:
#             return f"`pack.mcmeta` not found in pack {pack.filename}"

#     return False


def upload_flow(ui: DeltaGenerator):
    progress = ui.container()
    packs = ui.file_uploader("Upload packs", accept_multiple_files=True, type="zip")

    col1, col2, col3 = ui.columns(3)
    with col2:
        if fabric_mod := toggle_button("Turn into Fabric Mod", key="toggle"):
            ui.warning(webapp.fabric)

    path = None
    t0 = time.perf_counter()
    if col1.button("Build Packs", disabled=not packs, key="build") and packs:
        for pack in packs:
            with (temp / pack.name).with_suffix(".zip").open("wb") as temp_file:
                temp_file.write(pack.read())
        pack_paths = list(str(path) for path in temp.glob("*.zip"))
        # if error := validate_zips(pack_paths):
        #     st.error(error)
        #     return
        with st.status(f"Welding {len(packs)} packs!", expanded=True) as status:
            stream.seek(0)
            stream.truncate(0)
            with weld.run_weld(pack_paths, as_fabric_mod=fabric_mod) as ctx:
                if fabric_mod:
                    with ZipFile("output.jar", "w") as jar:
                        ctx.data.dump(jar)
                        ctx.assets.dump(jar)
                        path = Path(str(jar.filename))
                else:
                    if ctx.data:
                        path = ctx.data.save(path="output", zipped=True, overwrite=True)
                    if ctx.assets:
                        path = ctx.assets.save(
                            path="output", zipped=True, overwrite=True
                        )

            t1 = time.perf_counter()
            stream.seek(0)
            st.text_area("Build Log", stream.getvalue(), disabled=True, height=300)
            status.update(label=f"Merged in :green[{t1 - t0: 0.3f}s]")

        if path is not None:
            with path.open(mode="rb") as file:
                col3.download_button(
                    "Download Welded Packs",
                    file,
                    file_name="welded-pack.zip" if not fabric_mod else "welded-mod.jar",
                )
    else:
        col3.button("Download Welded Packs", disabled=True, key="download")

    if not packs:
        progress.info("Waiting for uploaded packs")
    elif path is None:
        progress.success("Ready to Build!")


def main():
    st.set_page_config(page_title="Weld", page_icon=icon, layout="wide")

    st.sidebar.title("Special Merging Rules")
    st.sidebar.write(webapp.conflicts)
    st.sidebar.write("----")
    st.sidebar.write("See [docs](https://wiki.smithed.dev/weld) for more info!")

    st.write(webapp.title.format(version=weld.__version__), unsafe_allow_html=True)

    st.write(webapp.intro, unsafe_allow_html=True)

    st.warning(webapp.warn)
    upload_flow(st.container())

    st.write("---")

    col1, col2 = st.columns(2)
    col1.markdown(webapp.footer.left, unsafe_allow_html=True)
    col2.markdown(webapp.footer.right, unsafe_allow_html=True)



import logging
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import click
import typer
from beet import BeetException, WrappedException
from beet.core.utils import format_exc

from .theming import print

logger = logging.getLogger("weld")


def path_str(path: Path) -> str:
    """Returns str of a path, replacing the home directory with '~'"""

    return str(path).replace(str(Path.home()), "~")


def format_error(
    message: str,
    exception: BaseException | None = None,
    padding: int = 0,
) -> str:
    """Format a given error message and exception."""
    output = "\n" * padding
    output += f"[bold][error]Error: {message}[/bold][/error]\n"
    if exception:
        output += "\n" + format_exc(exception)
    output += "\n" * padding
    return output.replace(str(Path.home()), "~")


@contextmanager
def error_handler(should_exit: bool = False, format_padding: int = 0) -> Iterator[None]:
    """Context manager that catches and displays exceptions."""
    exception = None

    try:
        yield
    except WrappedException as exc:
        message = str(exc)
        if not exc.hide_wrapped_exception:
            exception = exc.__cause__
    except BeetException as exc:
        message = str(exc)
    except (typer.Abort, KeyboardInterrupt):
        print()
        message = "Aborted."
    except (click.ClickException, click.exceptions.Exit):
        raise
    except Exception as exc:
        message = "An unhandled exception occurred. This could be a bug."
        exception = exc
    else:
        return

    logger.error(
        format_error(message, exception, format_padding), extra={"markup": True}
    )

    if should_exit:
        typer.Exit(-1)



"""
This file is super scrappy and needs to be rewritten but I can't be bothered right now.
"""

import logging
from dataclasses import dataclass
from typing import Any, Iterator, NamedTuple

from tokenstream import Token, TokenStream

from smithed.type import JsonDict, JsonList, JsonType

logger = logging.getLogger("weld")


@dataclass
class Filter:
    key: str
    value: Any
    type: str = "filter"


class TraverseResult(NamedTuple):
    parent: JsonDict | JsonList
    current: JsonType
    key: Any


def parse(raw: str, convert_index: bool):
    stream = TokenStream(raw)
    return handle_stream(stream, convert_index)


def handle_stream(stream: TokenStream, convert_index: bool) -> Iterator[Token | Filter]:
    with stream.syntax(
        key=r"[A-Z_a-z]+",
        string=r'"\w*"',
        separator=r"\.",
        bracket=r"\[|\]",
        curly=r"\{|\}",
        number=r"\d+",
        colon=r": *",
    ):
        for token in stream.collect():
            match token:
                case Token(type="key"):
                    yield token

                case Token(type="bracket", value="["):
                    match stream.expect_any("number", ("curly", "{")):
                        case Token(type="number", value=value) as token:
                            if convert_index:
                                yield Filter(key="_index", value=int(value))
                            else:
                                yield token
                        case Token(type="curly"):
                            yield parse_filter(stream)
                            stream.expect(("curly", "}"))
                    stream.expect(("bracket", "]"))

                case Token(type="separator") as token:
                    yield token


def parse_filter(stream: TokenStream):
    match stream.expect_any("key", "string"):
        case Token(type="key", value=value):
            key = value
        case Token(type="string", value=value):
            key = value[1:-1]
        case token:
            raise ValueError(f"Expected key or string. Found {token}")

    stream.expect("colon")
    return Filter(key=key, value=stream.expect_any("string", "number").value)


def traverse(obj: JsonDict, path: str, convert_index: bool = False):
    current = obj
    parent = current
    parser = parse(path, convert_index)
    last_token = None

    for token in parser:
        match token:
            case Token(type="key", value=value):
                if not isinstance(current, dict):
                    raise ValueError
                parent = current
                current = current.setdefault(value, {})
                last_token = token

            case Token(type="number", value=index):
                if not isinstance(current, list):
                    raise ValueError
                elif len(current) <= int(index):
                    raise IndexError
                parent = current
                current = current[int(index)]
                last_token = token

            case Filter(key=key, value=value):
                if not isinstance(current, list):
                    if last_token is None:
                        raise ValueError
                    if not isinstance(parent, dict):
                        raise ValueError
                    current = []
                    parent[last_token.value] = current  # type: ignore

                for item in current:
                    if not isinstance(item, dict):
                        raise ValueError
                    if key in item and item[key] == value:
                        parent = current
                        current = item
                        break
                else:
                    current.append({key: value})
                    parent = current
                    current = current[-1]
                last_token = token

    return TraverseResult(
        parent, current, last_token.value if last_token is not None else None
    )


def get(obj: JsonDict, path: str, convert_index: bool = False):
    return traverse(obj, path, convert_index).current


def append(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.append(value)
    elif not current:
        parent[key] = [value]


def prepend(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.insert(0, value)
    elif not current:
        parent[key] = [value]


def insert(obj: JsonDict, path: str, index: int, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.insert(index, value)
    elif not current:
        parent[key] = [value]


def remove(obj: JsonDict, path: str):
    parent, _, key = traverse(obj, path, True)
    if isinstance(parent, list):
        del parent[int(key)]
    else:
        del parent[key]


def merge(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)

    def _merge(
        parent: JsonDict | JsonList, original: JsonType, to_merge: JsonType, key: str
    ):
        if isinstance(original, dict) and isinstance(to_merge, dict):
            for new_key, val in to_merge.items():
                if og_val := original.get(new_key, False):
                    parent = original
                    _merge(parent, og_val, val, new_key)
                else:
                    original[new_key] = val

        elif isinstance(original, list) and isinstance(to_merge, list):
            original.extend(to_merge)

        else:
            # logger.warn("Deprecated. Use `smithed:replace` instead.")
            parent[key] = to_merge  # TODO: this might error !!

    _merge(parent, current, value, key)


def replace(obj: JsonDict, path: str, value: JsonType):
    parent, _, key = traverse(obj, path, True)
    parent[key] = value



from ..errors import WeldError


class MergingError(WeldError):
    """Error resulting from the merging process"""


class PriorityError(MergingError):
    """Error related to resolving priorities"""



from typing import Literal, Union

from .base import BaseModel

Condition = Union["ConditionPackCheck", "ConditionInverted"]


class ConditionPackCheck(BaseModel):
    type: Literal["pack_check", "weld:pack_check", "smithed:pack_check"]
    id: str


class ConditionInverted(BaseModel):
    type: Literal["inverted", "weld:inverted", "smithed:inverted"]
    conditions: list["Condition"]


ConditionInverted.update_forward_refs()



from typing import Annotated, Any, Literal

from pydantic import Field, validator

from .base import BaseModel


class _Source(BaseModel):
    type: str

    @validator("type")
    def fix_type(cls, value: str):
        if value.startswith("smithed:"):
            return value.replace("smithed:", "weld:")
        return value


class ReferenceSource(_Source):
    type: Literal["reference", "weld:reference", "smithed:reference"]
    path: str


class ValueSource(_Source):
    type: Literal["value", "weld:value", "smithed:value"] = "weld:value"
    value: Any


Source = Annotated[ValueSource | ReferenceSource, Field(..., discriminator="type")]



import sys
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Iterable, Literal, cast
from zipfile import Path as ZPath
from zipfile import ZipFile

from beet import Context, ProjectCache, ProjectConfig, run_beet, subproject
from beet.core.utils import FileSystemPath, JsonDict

from ..errors import WeldError
from .helper_plugins import add_fabric_mod_json

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

DESCRIPTION = "Merged by Smithed Weld"


class PackType(StrEnum):
    DATA = "data_pack"
    ASSETS = "resource_pack"


def subproject_config(pack_type: PackType, name: str = ""):
    return subproject(
        {
            "require": ["beet.contrib.auto_yaml"],
            pack_type: {"load": name},
            "pipeline": ["weld.print_pack_name", "weld.inject_pack_id_into_smithed"],
        }
    )


def inspect_zipfile(file: ZipFile) -> PackType:
    path = ZPath(file)

    if (path / "data").is_dir():
        return PackType.DATA
    elif (path / "assets").is_dir():
        return PackType.ASSETS

    raise WeldError("Invalid. Pack has neither assets nor data.")


def inspect(file: str | ZipFile) -> PackType | Literal[False]:
    match file:
        case str(path) if path.endswith(".zip"):
            with ZipFile(path) as zip:
                return inspect_zipfile(zip)

        case str(path):
            if (path := Path(path)).is_dir():
                if (path / "data").is_dir():
                    return PackType.DATA
                elif (path / "assets").is_dir():
                    return PackType.ASSETS

                raise WeldError(f"Invalid. Pack {path} has neither assets nor data. \n")

        case ZipFile() as zip:
            return inspect_zipfile(zip)

    return False


@contextmanager
def run_weld(
    packs: Iterable[str] | Iterable[ZipFile],
    config: FileSystemPath | ProjectConfig | JsonDict = {},
    directory: FileSystemPath | None = None,
    cache: bool | ProjectCache = True,
    as_fabric_mod: bool = False,
):
    packs = cast(list[str] | list[ZipFile], list(packs))
    packs_with_types = zip(packs, (pack for pack in map(inspect, packs) if pack))

    with run_beet(config, directory=directory, cache=cache) as ctx:
        ctx.require(weld)
        ctx.require(partial(load_packs, packs=list(packs_with_types)))
        ctx.require("weld.merging.process")
        if as_fabric_mod:
            ctx.require(partial(add_fabric_mod_json, packs=packs))
        yield ctx


def weld(ctx: Context):
    ctx.require("weld.merging")
    ctx.require("beet.contrib.model_merging")
    ctx.require("beet.contrib.unknown_files")

    # yield

    # ctx.require("weld.merging.process")


def load_packs(ctx: Context, packs: Iterable[tuple[str | ZipFile, PackType]]):
    for pack, pack_type in packs:
        match pack:
            case str(name):
                ctx.require(subproject_config(pack_type, name))
            case ZipFile() as file:
                match pack_type:
                    case PackType.DATA:
                        ctx.data.load(file)
                    case PackType.ASSETS:
                        ctx.assets.load(file)



from pydantic import BaseModel


class WebApp(BaseModel):
    class Footer(BaseModel):
        left: str
        right: str

    title: str
    conflicts: str
    intro: str
    warn: str
    fabric: str
    footer: Footer



from __future__ import annotations

import logging
from typing import Annotated, Literal

from pydantic import Field, validator

from .base import BaseModel
from .conditions import Condition
from .priority import Priority
from .sources import Source

logger = logging.getLogger(__name__)


class BaseRule(BaseModel):
    type: str
    target: str
    conditions: list[Condition] = []
    priority: Priority | None = None

    @validator("type")
    def fix_type(cls, value: str):
        if value.startswith("smithed:"):
            return value.replace("smithed:", "weld:")
        return value


class AdditiveRule(BaseRule):
    source: Source


class MergeRule(AdditiveRule):
    type: Literal["merge", "weld:merge", "smithed:merge"]


class AppendRule(AdditiveRule):
    type: Literal["append", "weld:append", "smithed:append"]


class PrependRule(AdditiveRule):
    type: Literal["prepend", "weld:prepend", "smithed:prepend"]


class InsertRule(AdditiveRule):
    type: Literal["insert", "weld:insert", "smithed:insert"]
    index: int


class ReplaceRule(AdditiveRule):
    type: Literal["replace", "weld:replace", "smithed:replace"]


class RemoveRule(BaseRule):
    type: Literal["remove", "weld:remove", "smithed:remove"]


Rule = Annotated[
    MergeRule | AppendRule | PrependRule | InsertRule | ReplaceRule | RemoveRule,
    Field(..., discriminator="type"),
]



from .conditions import Condition, ConditionInverted, ConditionPackCheck
from .main import SmithedJsonFile, SmithedModel, deserialize
from .priority import Priority
from .rules import (
    AdditiveRule,
    AppendRule,
    InsertRule,
    MergeRule,
    PrependRule,
    RemoveRule,
    ReplaceRule,
    Rule,
)
from .sources import ReferenceSource, Source, ValueSource

__all__ = [
    "deserialize",
    "AdditiveRule",
    "MergeRule",
    "AppendRule",
    "PrependRule",
    "InsertRule",
    "ReplaceRule",
    "RemoveRule",
    "Rule",
    "Condition",
    "ConditionInverted",
    "ConditionPackCheck",
    "Priority",
    "ReferenceSource",
    "ValueSource",
    "Source",
    "SmithedModel",
    "SmithedJsonFile",
]



from rich.console import Console
from rich.theme import Theme


def _brand():
    blue = accent = "#1B48C4"
    accent_light = "#23A3FF"
    locals()["bar.pulse"] = secondary = "#C41B9C"
    locals()["rule.line"] = green = success = "#1BC443"
    red = error = "#C41B1B"
    yellow = warning = "#E8AA03"

    white = foreground = "#FFF8F0"
    light_gray = border = "#4B4B4B"
    gray = highlight = "#2E2E31"
    dark_gray = section = "#1D1F21"
    background = "#121213"
    black = "#000000"

    return locals()


console = Console(theme=Theme(_brand()))
print = console.print



from beet import Context, LootTable

from .handler import ConflictsHandler


def beet_default(ctx: Context):
    handler = ctx.inject(ConflictsHandler)
    ctx.data.merge_policy.extend_namespace(LootTable, handler)


def process(ctx: Context):
    handler = ctx.inject(ConflictsHandler)
    handler.process()



import logging
from importlib import resources
from zipfile import ZipFile

from beet import Context, JsonFile, JsonFileBase
from jinja2 import Template

FABRIC_MOD_TEMPLATE = Template(
    (resources.files("weld") / "resources" / "fabric.mod.json.j2").read_text()
)

logger = logging.getLogger("weld")


def print_pack_name(ctx: Context):
    """Prints the merging pack name"""

    if not ctx.data and not ctx.assets:
        return

    logger.info(
        f"Loading {'data' if ctx.data else 'resource'}pack:"
        f" {ctx.data.name or ctx.assets.name or 'Unknown'}"
    )


def add_fabric_mod_json(ctx: Context, packs: list[str] | list[ZipFile]):
    ctx.data.extra["fabric.mod.json"] = JsonFile(
        FABRIC_MOD_TEMPLATE.render(
            packs=packs,
            pack_names=[
                pack.filename if type(pack) is ZipFile else pack for pack in packs
            ],
            mc_version=ctx.minecraft_version,
        )
    )


def inject_pack_id_into_smithed(ctx: Context):
    """CURSED"""

    id = (
        ctx.data.mcmeta.data.get("id", "missing")
        if ctx.data
        else ctx.assets.mcmeta.data.get("id", "missing")
    )

    for file in ctx.select(match="*", extend=JsonFileBase):
        try:
            if isinstance(file.data, list):  # type: ignore
                continue
            if smithed := file.data.get("__smithed__"):  # type: ignore
                if isinstance(smithed, list):
                    for item in smithed:
                        item["id"] = id
                elif isinstance(smithed, dict):
                    smithed["id"] = id
        except Exception:
            logger.warning(
                "Failed to inject pack id into smithed: %s", file.source_path
            )



"""This is the main logic for weld's custom merging logic.

It uses the beet's merge policies to implement a conflict handler that registers
 conflicts, merging all `__smithed__` into a base file (whether it's vanilla or not).
 In a later invocation, it then processes all of the registered conflicts, determines
 the ordering via the priority and stage system, and applies each defined rule onto
 the base file.

 TODO: Likely refactor into multiple files, I can see each file handling being it's own
  class since several methods based on each file are passing similar parameters with
  each other.
"""

import json
import logging
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass, field
from importlib import resources
from typing import Literal, cast

from beet import Context, DataPack, JsonFile, NamespaceFile
from beet.contrib.format_json import get_formatter
from beet.contrib.vanilla import Vanilla
from pydantic import ValidationError

from smithed.type import JsonDict, JsonTypeT

from ..models import (
    AppendRule,
    Condition,
    ConditionInverted,
    ConditionPackCheck,
    InsertRule,
    MergeRule,
    PrependRule,
    RemoveRule,
    ReplaceRule,
    Rule,
    SmithedJsonFile,
    ValueSource,
    deserialize,
)
from .errors import PriorityError
from .parser import append, get, insert, merge, prepend, remove, replace

logger = logging.getLogger("weld")

PRIORITY_STAGES = ["early", "standard", "late"]
YELLOW_SHULKER_BOX = resources.files("weld") / "resources" / "yellow_shulker_box.json"


@dataclass
class ConflictsHandler:
    ctx: Context

    formatter: Callable[..., str] = get_formatter()
    cache: defaultdict[type[NamespaceFile], set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )
    vanilla: set[str] = field(default_factory=set)
    overrides: set[str] = field(default_factory=set)

    def __call__(
        self, pack: DataPack, path: str, current: JsonFile, conflict: JsonFile, /
    ):
        """Register conflicts.."""

        logger.debug(f"Registering conflict: {path!r}")
        if path in self.overrides:
            logger.debug("Skipping due to override")
            return True

        # Parse the files and handle validation errors. We need to ensure that `current`
        #  is left with a valid file. If a file has an incorrect `smithed` definition.
        smithed_current = self.parse_smithed_file(current)
        smithed_conflict = self.parse_smithed_file(conflict)

        if smithed_current is False and smithed_conflict is False:
            logger.warn(
                "Both the current and conflict files are invalid. Undefined Behavior"
            )
            return False

        elif smithed_current is False:
            current.data = conflict.data
            return True

        elif smithed_conflict is False:
            return True

        current_entries = smithed_current.smithed.entries()

        if len(current_entries) > 0 and current_entries[0].override:
            logger.critical(
                f"Overriding base file at `{path}` with {current_entries[0].id}"
            )
            self.overrides.add(path)
            return True

        conflict_entries = smithed_conflict.smithed.entries()
        if len(conflict_entries) > 0 and conflict_entries[0].override:
            logger.critical(
                f"Overriding base file at `{path}` with {conflict_entries[0].id}"
            )
            self.overrides.add(path)
            current.data = conflict.data
            return True

        # Cache paths for latest use
        json_file_type = cast(type[NamespaceFile], type(current))
        self.cache[json_file_type].add(path)

        # Handle vanilla paths as the base / current file
        if path.startswith("minecraft:"):
            if path not in self.vanilla:
                current.data = self.grab_vanilla(path, json_file_type)
                if path == "minecraft:blocks/yellow_shulker_box":
                    logger.warn(
                        f"Weld is ignoring [bold]'{path}'[/bold] for now.",
                        extra={"markup": True},
                    )
                    return True

                self.vanilla.add(path)

        # Handle non-vanilla paths, swap conflict w/ current if no smithed rules exist
        # This is to ensure that non-vanilla files can work with weld.
        # Example:
        #   TCC has a loot table for iceologer
        #   Pack B wants to weld the loot table for iceologer
        #   This will ensure that TCC is the base reference even if Pack B is
        #    loaded first as TCC will not have any weld rules defined but Pack B will.
        #
        # ⚠️ It's important that either the current or conflict files have smithed rules
        #  though it be odd if two packs are writing to the same namespace.
        elif not smithed_conflict.smithed.entries():
            if not current_entries:
                if smithed_conflict.dict() != smithed_current.dict():
                    logger.warn(
                        f"Conflict unresolved at '{path}'.\nContents are different and"
                        f" contain no smithed rules which is likely unintended."
                    )
            else:
                logger.info(f"Swapping base file at `{path}`")
                current.data, conflict.data = conflict.data, current.data
                smithed_current, smithed_conflict = smithed_conflict, smithed_current

        # Grab conflicts to throw with the current file
        if conflict_entries := smithed_conflict.smithed.entries():
            current_entries.extend(conflict_entries)

        # Save back to current file
        raw: JsonDict = deserialize(smithed_current)
        current.data["__smithed__"] = raw["__smithed__"]

        return True

    def parse_smithed_file(self, file: JsonFile) -> SmithedJsonFile | Literal[False]:
        """Parses a smithed file and returns the parsed file or False if invalid."""

        try:
            return SmithedJsonFile.parse_obj(file.data)
        except ValidationError:
            logger.error("Failed to parse smithed file ", exc_info=True)
            return False

    def grab_vanilla(self, path: str, json_file_type: type[NamespaceFile]) -> JsonDict:
        """Grabs the vanilla file to load as the current file (aka the base).

        ⚠️ Uses the bundled `yellow_shulker_box.json` as an override over vanilla's.

        """
        if path == "minecraft:blocks/yellow_shulker_box":
            return json.loads(YELLOW_SHULKER_BOX.read_text())

        vanilla = self.ctx.inject(Vanilla)
        return cast(JsonFile, vanilla.data[json_file_type][path]).data

    def process(self):
        """Main entrypoint for smithed merge solving"""

        for json_file_type, path in self:
            logger.info(f"Resolving {json_file_type.__name__}: {path!r}")

            namespace_file = self.ctx.data[json_file_type]
            smithed_file = SmithedJsonFile.parse_obj(
                namespace_file[path].data  # type: ignore
            )

            if smithed_file.smithed.entries():
                processed = self.process_file(smithed_file)

                # reorder so `__smithed__` is at the bottom in output
                temp = processed["__smithed__"]
                del processed["__smithed__"]
                processed["__smithed__"] = temp

                namespace_file[path].data = processed  # type: ignore

    def process_file(self, file: SmithedJsonFile) -> JsonDict:
        """Process each file's rules"""

        logger.debug("Resolving priorities")
        rules = self.resolve_priorities(file)
        logger.debug(
            "Rules: %s",
            ", ".join(f"{id} {type(rule).__name__}" for id, rule in rules.items()),
        )

        logger.debug("Injecting `_index`")
        raw = self.manage_indexes(deserialize(file, defaults=False))

        logger.debug(f"Pack order: {', '.join(rules)}")
        for id, rule in rules.items():
            if applied := self.apply_rule(raw, id, rule):
                raw = applied

        return self.manage_indexes(raw, strip=True)

    def manage_indexes(self, data: JsonTypeT, strip: bool = False) -> JsonTypeT:
        """Adds / removes `_index` field to every item in a list"""

        match data:
            case list(value):
                for index, item in enumerate(value):
                    if isinstance(item, dict):
                        if strip:
                            item.pop("_index", None)
                        else:
                            item["_index"] = index

                return [
                    self.manage_indexes(item, strip) for item in value  # type: ignore
                ]

            case dict(value):
                return {
                    key: self.manage_indexes(val, strip)  # type: ignore
                    for key, val in value.items()
                }

            case other:
                return other

    def pre_process_condition(
        self, rules: dict[str, Rule], condition: Condition
    ) -> bool:
        """Returns true if all conditions satisfy their criteria."""

        match condition:
            case ConditionPackCheck(id=id):
                return id in rules
            case ConditionInverted(conditions=conditions):
                return not all(
                    self.pre_process_condition(rules, condition)
                    for condition in conditions
                )

    def pre_process_rules(self, file: SmithedJsonFile):
        """Gathers models by id. Pre-processes models for ease of use.

        Converts each 'before' priority into 'after' (as it's easier to resolve).
        """

        # refactor to be cleaner
        rules = {
            model.id: rule for model in file.smithed.entries() for rule in model.rules
        }
        removed_ids: set[str] = set()
        for current_id, rule in rules.items():
            if not all(
                self.pre_process_condition(rules, condition)
                for condition in rule.conditions
            ):
                logging.info("Skipping rule from %s due to conditions", current_id)
                removed_ids.add(current_id)
                continue

            assert rule.priority is not None
            if before := rule.priority.before.entries():
                for id in before:
                    if id not in rules:
                        logger.warn(f"Priority: {id} was not found. Ignoring Rule.")
                        continue
                    other_rule = rules[id]
                    assert other_rule.priority is not None
                    if current_id not in other_rule.priority.after.entries():
                        other_rule.priority.after.entries().append(current_id)
                before.clear()

        return {id: rule for id, rule in rules.items() if id not in removed_ids}

    def resolve_priorities(self, file: SmithedJsonFile):
        """Resolves priorities for each stage."""

        pre_processed_rules = self.pre_process_rules(file)
        rules: dict[str, Rule] = {}

        for stage in PRIORITY_STAGES:
            logger.debug(f"Stage: `{stage}`")
            for id, rule in pre_processed_rules.items():
                assert rule.priority is not None
                if rule.priority.stage == stage:
                    self.resolve_stage(stage, pre_processed_rules, id, rules, [])

        return rules

    def resolve_stage(
        self,
        stage: str,
        pre_processed_rules: dict[str, Rule],
        current: str,
        processed: dict[str, Rule],
        processing: list[str],
    ):
        """Resolves priorities within a specific stage.

        ⚠️ Raises PriorityError if a dependency loop is found.
        ⚠️ Raises PriorityError if a pack depends on another pack from a different stage.
        ⚠️ Logs warning if a pack depends on a pack that does not exist.
        """

        current_rule = pre_processed_rules[current]

        assert current_rule.priority is not None
        if after := current_rule.priority.after.entries():
            for id in after:
                if id in processing:
                    raise PriorityError(
                        "Dependency loop found while resolving"
                        f" `{current}` in stage `{stage}`.\n"
                        "The following loop was discovered: "
                        + " -> ".join(f"`{id}`" for id in processing)
                        + f" -> `{id}`"
                    )
                if stage != current_rule.priority.stage and id not in processed:
                    raise PriorityError(
                        f"Cannot resolve priority for rule during stage `{stage}`."
                        "\nPacks must only depend on packs in `before` or `after`"
                        " thah are in the same stage. You likely shouldn't specify"
                        " both `stage` AND `before/after` unless you know what"
                        " you are doing."
                    )
                elif id in pre_processed_rules:
                    processing.append(id)
                    self.resolve_stage(
                        stage, pre_processed_rules, id, processed, processing
                    )
                    processing.remove(id)
                else:
                    logger.warn(f"Priority: {id} was not found. Ignoring Rule.")

        # dict insertion order for the win
        if current not in processed:
            processed[current] = pre_processed_rules[current]

    def apply_rule(
        self, raw: JsonDict, current: str, rule: Rule
    ) -> JsonDict | Literal[False]:
        """Rule application uses `jsonpath_ng` for parsing target paths.

        TODO: extract outward, maybe bundle rule application with the rule itself
        """

        # Handle whether the target path exists or not
        try:
            get(raw, rule.target, True)
        except ValueError:
            logger.warn(
                f"Target Path: {rule.target} was not found. Ignoring...", exc_info=True
            )
            return False

        # Apply each rule's logic
        match rule:
            case MergeRule(source=ValueSource(value=value)):
                merge(raw, rule.target, value)

            case AppendRule(source=ValueSource(value=value)):
                append(raw, rule.target, value)

            case PrependRule(source=ValueSource(value=value)):
                prepend(raw, rule.target, value)

            case InsertRule(source=ValueSource(value=value), index=index):
                insert(raw, rule.target, index, value)

            case ReplaceRule(source=ValueSource(value=value)):
                replace(raw, rule.target, value)

            case RemoveRule():
                remove(raw, rule.target)

        return raw

    def __iter__(self):
        for json_file_type, paths in self.cache.items():
            yield from [(json_file_type, path) for path in paths]
